# MiroTalk SFU Integration for Drupal

The MiroTalk SFU module enables seamless integration of video calling and
screen-sharing functionalities in Drupal by leveraging the MiroTalk SFU server.
It provides a simple interface for managing video call rooms, connecting to
the MiroTalk SFU server, and customizing video call settings.

## Configuration

Before using the module, ensure that you have configured the following fields:

- `Server Path`: The path to the server's MiroTalk SFU source,
- `Video Storage Path`: The URL for storing video recordings.
- `Config server`: Set up server management information

## Features

- Integration with MiroTalk SFU for video calls and screen sharing.
- Create and manage video call rooms directly from your Drupal site.
- Built-in support for recording and storing video calls.
- Redirect users to a specified URL after leaving the call.

## Requirements
- A linux server running (Ubuntu 20.04 LTS or later).
- Drupal: Version 10.x or higher.
- MiroTalk SFU Server: Fully configured & operational MiroTalk SFU.
- Browser: A modern browser with WebRTC support (e.g., Chrome, Firefox).

## Installation
1. Download & install the MiroTalk SFU server [https://github.com/miroslavpejic85/mirotalksfu](https://github.com/miroslavpejic85/mirotalksfu).
2. High recommend download to /opt (ubuntu)
3. Go to dir mirotalksfu: `cd mirotalksfu`
4. Copy app/src/config.template.js in app/src/config.js and edit it:
`cp app/src/config.template.js app/src/config.js`
5. Copy .env.template to .env and edit it if needed: `cp .env.template .env`
6. open permission ownership www-data for .env `chown www-data .evn`
7. Install dependencies: `npm install`
8. Start the server: `npm start`

Example setup on Ubuntu 24.04
<pre>
sudo apt install -y python3.8 python3-pip ffmpeg nodejs
sudo npm install -g pm2 dotenv
cd /opt
sudo git clone https://github.com/miroslavpejic85/mirotalksfu.git
cd mirotalksfu
cp app/src/config.template.js app/src/config.js
cp .env.template .env
chown www-data:www-data .evn

sudo npm install
sudo nano ecosystem.config.js
</pre>

Example
#/opt/mirotalksfu/ecosystem.config.js
<pre>
module.exports = {
apps: [
{
name: "Miro Talk",
script: "./app/src/Server.js",
        // Env Specific Config
        env_production: {
          NODE_ENV: "production",
          PORT: 3000,
          exec_mode: "cluster_mode",
        },
      },
    ],
};
</pre>

<pre>
cd /etc/apache2/sites-available/
sudo nano mirotalk.yoursite.com.conf
</pre>

Content of mirotalk.yoursite.com.conf
<code>
<VirtualHost *:80>
    ServerName mirotalk.yoursite.com
    Redirect permanent / https://mirotalk.yoursite.com/
    RewriteEngine on
    RewriteCond %{SERVER_NAME} =mirotalk.yoursite.com
    RewriteRule ^ https://%{SERVER_NAME}%{REQUEST_URI} [END,NE,R=permanent]
</VirtualHost>
<VirtualHost *:443>
    ServerName mirotalk.yoursite.com
    # Config Proxy
    ProxyPass / http://localhost:3000/
    ProxyPassReverse / http://localhost:3000/
    # Config Proxy WebSocket
    RewriteEngine On
    RewriteCond %{HTTP:Upgrade} =websocket [NC]
    RewriteCond %{REQUEST_URI} ^/socket.io/ [NC]
    RewriteRule /(.*) ws://127.0.0.1:3000/$1 [P,L]
    # Config Header
    Header always set X-Content-Type-Options "nosniff"
    Header always set X-Frame-Options "DENY"
    Header always set Strict-Transport-Security "max-age=31536000;
includeSubDomains;preload"
    # Log file
    ErrorLog ${APACHE_LOG_DIR}/mirotalk_error.log
    CustomLog ${APACHE_LOG_DIR}/mirotalk_access.log combined
</VirtualHost>
</code>

<pre>
sudo a2enmod proxy proxy_http headers
sudo a2ensite mirotalk.yoursite.com
sudo systemctl restart apache2
pm2 start /opt/mirotalksfu/ecosystem.config.js --env production
</pre>

## Important
1. The module must have write permission for the .env file at
the root of the installed mirotalksfu, example:
2. `cd /opt/mirotalksfu`
3. `sudo cp .env.template .env`
4. `sudo chown www-data:www-data .env`

## Setup
1. Go to `/admin/config/services/mirotalk-sfu-meeting` to configure
Miro talk server.
2. After install module MiroTalk SFU you will have new entity type
3. You can create new meeting room, or you can add like field reference
in your content type


## Usage

Once installed and configured, you can create video call rooms
and invite users to join. The module will handle the connection
to the MiroTalk SFU server, allow for screen sharing, and provide
functionality for managing video call sessions.

## Contributions

Feel free to submit issues and pull requests for improvements.
